//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
#include "stdafx.h"

#include "QmolDoc.h"
#include "AdjustTorsion.h"
#include <math.h>

// Local function definition
BOOL GetRotationMatrix(float mat[9], float angle, 
					   float x, float y, float z);

#define		DEG2RAD		(1.74532925e-2) // Pi/180.0

// Recursively find all of the atoms connected to atom "a" except for
// atoms listed in the atom_buffer. Atoms are identified by
// index (not serial) number.
void AdjustTorsion::Trace(array< array<unsigned int> > &m_molConnect, 
						  int a, array<int> &atom_buffer)
{
	if(atom_buffer[a])
		return;
	
	atom_buffer[a] = 1;
	
	for(int i = 0;i < m_molConnect[a].Length();i++)
		Trace(m_molConnect, m_molConnect[a][i], atom_buffer);
}

// Find the smalled set of atoms whose position is affected by a
// change in the torsion angle about atoms "a" & "b".
void AdjustTorsion::Set(SelectObj &m_obj)
{
	int a = m_obj.GetMol()->bonds[m_obj.GetIndex()].a;
	int b = m_obj.GetMol()->bonds[m_obj.GetIndex()].b;

	// If the selected torsion angle is allready the current torsion angle,
	// then deselect it!
	if(m_obj.GetMol() == torMol){
		if(((tor_a == a) && (tor_b == b)) || ((tor_a == b) && (tor_b == a))){
			Init();
			return;
		}
	}
	
	torMol = m_obj.GetMol();

	// Clean up old memory as needed
	affected_atoms.Alloc(0);
	
	// Start with a clean slate
	last_tor = 0.0;

	int i;
	array<int> a_buf(torMol->NumAtoms());
	array<int> b_buf(torMol->NumAtoms());
	int num_a, num_b;

	// Save the torsion angle
	tor_a = a;
	tor_b = b;
		
	// Start with a clean slate
	a_buf = 0;
	b_buf = 0;
	num_a = 0;
	num_b = 0;
	
	a_buf[b] = 1;
	b_buf[a] = 1;
	
	Trace(torMol->mol_connect, a, a_buf);
	Trace(torMol->mol_connect, b, b_buf);
	
	// Neither a nor b are changed by the 
	// torsion angle
	a_buf[a] = 0;
	a_buf[b] = 0;
	b_buf[a] = 0;
	b_buf[b] = 0;
	
	
	int are_identical = 1;
	
	for(i = 0;i < a_buf.Length();i++){
		if(a_buf[i] != b_buf[i])
			are_identical = 0;
			
		num_a += a_buf[i];
		num_b += b_buf[i];
	}
		
	// If the two selection buffers are identical, the
	// user has selected a closed ring structure - so
	// don't return any atoms.
	if(are_identical){
		//affected_atoms.Alloc(0);
		return;
	}
	
	if(num_a <= num_b){
		affected_atoms.Alloc(num_a);
		num_a = 0;
		for(i = 0;i < a_buf.Length();i++)
			if(a_buf[i]){
				affected_atoms[num_a++] = i;
			}
	}
	else{
		affected_atoms.Alloc(num_b);
		num_b = 0;
		for(i = 0;i < b_buf.Length();i++)
			if(b_buf[i]){
				affected_atoms[num_b++] = i;
			}
	}
}

void AdjustTorsion::Update(float new_tor)
{
	ASSERT(torMol != NULL);

	float dtor = new_tor - last_tor;
	float r[3];
	float m[9]; // A matrix in "flattened" array form
	int i;
	
	if(torMol->current_step == 0){
		// Rotate about the vector r.
		// Work in a four dimensional system to take advantage
		// of the openGL rotation matricies.
		r[0] = torMol->atoms[tor_b].coord[0] - torMol->atoms[tor_a].coord[0];
		r[1] = torMol->atoms[tor_b].coord[1] - torMol->atoms[tor_a].coord[1];
		r[2] = torMol->atoms[tor_b].coord[2] - torMol->atoms[tor_a].coord[2];
		
		if(GetRotationMatrix(m, dtor, r[0], r[1], r[2]) == FALSE)
			// No rotation is needed (or the matrix could not be constructed).
			return;

		// Now rotate all of the atoms
		for(i = 0;i < affected_atoms.Length();i++){
			r[0] = torMol->atoms[affected_atoms[i]].coord[0] - 
				torMol->atoms[tor_a].coord[0];
			r[1] = torMol->atoms[affected_atoms[i]].coord[1] - 
				torMol->atoms[tor_a].coord[1];
			r[2] = torMol->atoms[affected_atoms[i]].coord[2] - 
				torMol->atoms[tor_a].coord[2];
			
			torMol->atoms[affected_atoms[i]].coord[0] = torMol->atoms[tor_a].coord[0] + 
				m[0]*r[0] + m[1]*r[1] + m[2]*r[2];
			torMol->atoms[affected_atoms[i]].coord[1] = torMol->atoms[tor_a].coord[1] + 
				m[3]*r[0] + m[4]*r[1] + m[5]*r[2];
			torMol->atoms[affected_atoms[i]].coord[2] = torMol->atoms[tor_a].coord[2] + 
				m[6]*r[0] + m[7]*r[1] + m[8]*r[2];
		}
	}
	else{
		int index = (torMol->current_step - 1);

		// Rotate about the vector r
		r[0] = torMol->traj[index][3*tor_b] - torMol->traj[index][3*tor_a];
		r[1] = torMol->traj[index][3*tor_b + 1] - torMol->traj[index][3*tor_a + 1];
		r[2] = torMol->traj[index][3*tor_b + 2] - torMol->traj[index][3*tor_a + 2];
		
		if(GetRotationMatrix(m, dtor, r[0], r[1], r[2]) == FALSE)
			// No rotation is needed (or the matrix could not be constructed).
			return;
		
		// Now rotate all of the atoms
		for(i = 0;i < affected_atoms.Length();i++){
			r[0] = torMol->traj[index][3*affected_atoms[i]] - 
				torMol->traj[index][3*tor_a];
			r[1] = torMol->traj[index][3*affected_atoms[i] + 1] - 
				torMol->traj[index][3*tor_a + 1];
			r[2] = torMol->traj[index][3*affected_atoms[i] + 2] - 
				torMol->traj[index][3*tor_a + 2];

			torMol->traj[index][3*affected_atoms[i]] = torMol->traj[index][3*tor_a] + 
				m[0]*r[0] + m[1]*r[1] + m[2]*r[2];
			torMol->traj[index][3*affected_atoms[i] + 1] = torMol->traj[index][3*tor_a + 1] + 
				m[3]*r[0] + m[4]*r[1] + m[5]*r[2];
			torMol->traj[index][3*affected_atoms[i] + 2] = torMol->traj[index][3*tor_a + 2] + 
				m[6]*r[0] + m[7]*r[1] + m[8]*r[2];

		}
	}
	
	// All done!
	last_tor = new_tor;
}

// This function is based on the _math_matrix_rotate function from
// the Mesa library (by Erich Boleyn, erich@uruk.org).
BOOL GetRotationMatrix(float mat[9], float angle, float x, float y, float z)
{
	float mag, s, c;
	float xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c;

	s = (float)(sin(angle*DEG2RAD));
	c = (float)(cos(angle*DEG2RAD));

	mag = (float)(sqrt(x*x + y*y + z*z));

	if (mag < PRECISION) {
	   // The matrix could not be constructed
      return FALSE;
	}
	
	mag = 1.0f/mag;

	x *= mag;
	y *= mag;
	z *= mag;

	one_c = 1.0f - c;

	xx = x * x * one_c;
	yy = y * y * one_c;
	zz = z * z * one_c;
	xy = x * y * one_c;
	yz = y * z * one_c;
	zx = z * x * one_c;

	xs = x * s;
	ys = y * s;
	zs = z * s;

	mat[0] = xx + c;
	mat[1] = xy - zs;
	mat[2] = zx + ys;

	mat[3] = xy + zs;
	mat[4] = yy + c;
	mat[5] = yz - xs;

	mat[6] = zx - ys;
	mat[7] = yz + xs;
	mat[8] = zz + c;

	return TRUE;
}
