//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
// 
#include "stdafx.h"
#include "CommandLine.h"
#include "Excepts.h"
#include "io.h"

// Current flags and arguments that are recognized:
// [filename]				Open coordinate file (type based on file extension)
// -pdb [filename]			Read PDB coordinate file
// -dcd [filename]			Read DCD trajectory file
// -align					Align all trajectory structures by rotation and translation
// -cm						Remove the center of mass from all trajectory structures
// -no_connect				Don't compute covalent bonds from atomic coordinates
// -multi					Interpret multiple structures in a pdb file as a trajectory
// -bpp	[number]			Override the default color depth
// -zbuf [number]			Override the default Z buffer size
// -?						Show the list of allowed command line options
// -help					Show the list of allowed command line options

// Helper function(s)
CString GetWildcardPathInfo(LPCTSTR lpszParam);

void CQmolCommandLine::ParseParam(LPCTSTR lpszParam, BOOL bFlag, BOOL bLast)
{
	// Take care of the base class
	//CCommandLineInfo::ParseParam(lpszParam, bFlag, bLast);
	
	// Do we expect that the current lpszParam contains
	// a filename for data import or parameter value?
	static BOOL read_coord_file_name = FALSE;
	static BOOL read_traj_file_name = FALSE;
	static BOOL read_bpp = FALSE;
	static BOOL read_zbuf = FALSE;

	// Have we read a flag?
	if(bFlag){
		if(read_coord_file_name || read_traj_file_name){
			m_importPDB = FALSE;
			m_importDCD = FALSE;
			throw QMOL_ERROR("A filename was not found on the command line");
		}
		
		if(strcmp(lpszParam, "align") == 0){
			m_traj_kabsch = TRUE;
		}
		
		if(strcmp(lpszParam, "cm") == 0){
			m_traj_cm = TRUE;
		}
		
		if(strcmp(lpszParam, "no_connect") == 0){
			m_coord_connect_on_read = FALSE;
		}

		if(strcmp(lpszParam, "multi") == 0){
			m_multi_traj = TRUE;
		}

		if(strcmp(lpszParam, "?") == 0){
			m_help = TRUE;
		}

		if(strcmp(lpszParam, "help") == 0){
			m_help = TRUE;
		}

		if(strcmp(lpszParam, "pdb") == 0){
			// The next item we parse must be the filename
			if(bLast){
				m_importPDB = FALSE;
				throw QMOL_ERROR("A filename was not found on the command line");
			}
			else{
				m_importPDB = TRUE;
				read_coord_file_name = TRUE;
			}
		}

		if(strcmp(lpszParam, "dcd") == 0){
			// The next item we parse must be the filename
			if(bLast){
				m_importDCD = FALSE;
				throw QMOL_ERROR("A filename was not found on the command line");
			}
			else{
				m_importDCD = TRUE;
				read_traj_file_name = TRUE;
			}
		}

		if(strcmp(lpszParam, "bpp") == 0){
			// The next item we parse must be the value
			if(bLast){
				m_set_bpp = FALSE;
				throw QMOL_ERROR("A bpp value was not found on the command line");
			}
			else{
				m_set_bpp = TRUE;
				read_bpp = TRUE;
			}
		}

		if(strcmp(lpszParam, "zbuf") == 0){
			// The next item we parse must be the value
			if(bLast){
				m_set_zbuf = FALSE;
				throw QMOL_ERROR("A zbuf value was not found on the command line");
			}
			else{
				m_set_zbuf = TRUE;
				read_zbuf = TRUE;
			}
		}
	}
	else{
		if(read_coord_file_name || read_traj_file_name || read_zbuf || read_bpp){

			if(read_coord_file_name){
				// Does this name contain a wild card?
				if(strchr(lpszParam, '*') != NULL){
					// This string contains a wild card, save the file name(s).
					struct _finddata_t fileinfo;
					long fileHandle;

					fileHandle = _findfirst(lpszParam, &fileinfo);

					if(fileHandle == -1){
						throw QmolError("The wildcard failed to match a file name", 
							__FILE__, __LINE__);
					}
					
					CString pathInfo = GetWildcardPathInfo(lpszParam);

					// Iterate over all file names that match the wildcard pattern while
					// making sure to append any path information.
					while(_findnext(fileHandle, &fileinfo) == 0){
						m_import_coord_filename.Push(pathInfo + CString(fileinfo.name));
					}

					_findclose(fileHandle);
				}
				else{
					// Save the file name
					m_import_coord_filename.Push(CString(lpszParam));
				}
			
				read_coord_file_name = FALSE;
			}

			if(read_traj_file_name){
				read_traj_file_name = FALSE;
			
				// Save the file name
				m_import_traj_filename = lpszParam;
			}

			if(read_bpp){
				read_bpp = FALSE;
			
				// Save the bpp
				m_bpp = atoi(lpszParam);
			}

			if(read_zbuf){
				read_zbuf = FALSE;
			
				// Save the zbuf
				m_zbuf = atoi(lpszParam);
			}
		}
		else{
			// The user has specified a coordinate file to open
			// (based on file extension).
			
			m_importAUTO = TRUE;
			
			if(strchr(lpszParam, '*') != NULL){
				// This string contains a wild card, save the file name(s).
				struct _finddata_t fileinfo;
				long fileHandle;
				
				fileHandle = _findfirst(lpszParam, &fileinfo);
				
				if(fileHandle == -1){
					throw QmolError("The wildcard failed to match a file name", 
						__FILE__, __LINE__);
				}
				
				CString pathInfo = GetWildcardPathInfo(lpszParam);
				
				// Iterate over all file names that match the wildcard pattern while
				// making sure to append any path information.
				while(_findnext(fileHandle, &fileinfo) == 0){
					m_import_coord_filename.Push(pathInfo + CString(fileinfo.name));
				}
				
				_findclose(fileHandle);
			}
			else{
				// Save the file name
				m_import_coord_filename.Push(CString(lpszParam));
			}
		}
	}
}

CString GetWildcardPathInfo(LPCTSTR lpszParam)
{
	CString pathinfo = _T("");

	// First identify the location of the wildcard character
	char *ptr = strchr(lpszParam, '*');

	if(ptr == NULL)
		return pathinfo;

	int len = 0;

	// Now search backwards for the location of the first '\' character
	while(ptr != lpszParam){
		if(*ptr == '\\'){
			len = ptr - lpszParam + 1;
			char *tmpString = new char [len + 1];
			strncpy(tmpString, lpszParam, len);
			tmpString[len] = '\0';

			pathinfo = tmpString;

			return pathinfo;
		}

		ptr -= 1;
	}
	
	return pathinfo;
}
