//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

#ifndef __QMOL_DEFINE
#define __QMOL_DEFINE

// Number of significant figures to use when saving data
#define	QMOL_DEFAULT_PRECISION	7

// Enumerate the different file types
enum {QM_PDB = 0, /* Coordiante files start here! */
	  QM_XYZ, 
	  QM_CRD, 
	  QM_DCD, /* Trajectory files start here! */
	  QM_PTF, 
	  QM_AUTO, /* General (i.e. both coord and trajec) types here! */
	  QM_UNK};

// Enumerate element type for hash lists (for element name, radius and color).
// ELEM_LAST is merely a place-hold for array sizing.
enum {ELEM_C = 0,				// Carbon
	  ELEM_N,					// Nitrogen
	  ELEM_O,					// Oxygen
	  ELEM_S,					// Sulphur
	  ELEM_H,					// Hydrogen
	  ELEM_P,					// Phosphorus
	  ELEM_CL,					// Chlorine
	  ELEM_Z,					// Zinc
	  ELEM_NA,					// Sodium
	  ELEM_FE,					// Iron
	  ELEM_MG,					// Magnesium
	  ELEM_K,					// Potassium
	  ELEM_CA,					// Calcium
	  ELEM_I,					// Iodine
	  ELEM_DUMMY,				// Unknown
	  ELEM_LAST};

// Enumerate the different amino acids
enum {
	AA_ALA = 0,
	AA_VAL,
	AA_PHE,
	AA_PRO,
	AA_MET,
	AA_ILE,
	AA_LEU,
	AA_ASP,
	AA_GLU,
	AA_LYS,
	AA_ARG,
	AA_SER,
	AA_THR,
	AA_TYR,
	AA_HIS,
	AA_CYS,
	AA_ASN,
	AA_GLN,
	AA_TRP,
	AA_GLY,
	AA_DUMMY,			// Unknown
	AA_LAST
};

// Enumerate the different secondary structure types (using the same
// definitions as the Stride program).
enum {
	SS_ALPHA_HELIX = 0,	// "AlphaHelix"
	SS_310_HELIX,		// "310Helix"
	SS_PI_HELIX,		// "PiHelix"
	SS_STRAND,			// "Strand" i.e. \beta sheet
	SS_BRIDGE,			// "Bridge"
	SS_COIL,			// "Coil"
	SS_TURN,			// "Turn"
	SS_TURN_1,			// "TurnI"
	SS_TURN_1P,			// "TurnI'"
	SS_TURN_2,			// "TurnII"
	SS_TURN_2P,			// "TurnII'"
	SS_TURN_4,			// "TurnIV"
	SS_TURN_6A,			// "TurnVIa"
	SS_TURN_6B,			// "TurnVIb"
	SS_TURN_8,			// "TurnVIII"
	SS_TURN_GAMMA,		// "GammaClassic"
	SS_TURN_GAMMA_INV,	// "GammaInv"
	SS_DUMMY,			// "Unknown"
	SS_LAST
};

// Some constants for computing internal variables
#ifdef PI   // Avoid Linux warnings
#undef PI
#endif

#define PI     3.141592653589793f

#define PRECISION   1.0e-7f

// Different processing options for reading trajectory files
enum {QM_VERBATIM = 0, QM_CENTER_ALL, QM_KABSCH};

// Enumerate the various color schemes
enum {QM_COLOR_DEFAULT, QM_COLOR_CPK, QM_COLOR_STRUC_DEFAULT,
	QM_COLOR_AMINO};

// Enumerate the various atom radius schemes
typedef enum {QM_RADIUS_DEFAULT = 0, QM_RADIUS_ELEMENT, QM_RADIUS_CUSTOM, QM_RADIUS_T,
QM_RADIUS_OCC} RADIUS_FORMAT;

// Different ways to display a molecule
typedef enum {QM_RENDER_NONE,			// Don't render the molecule!
			  QM_RENDER_WIREFRAME,		// Render the covalent bonds as lines
			  QM_RENDER_CPK,			// Render the atoms as spheres
			  QM_RENDER_POINT,			// Render the atoms as points
			  QM_RENDER_STICK,			// Renders the bonds as cylinders
			  QM_RENDER_BALL_AND_STICK, // Render the atoms as spheres and the 
									    // bonds as cylinders
			  QM_RENDER_TRACE_STICK,	// Render the C_\alpha trace with cylinders
			  QM_RENDER_TRACE_WIREFRAME,// Render the C_\alpha trace with lines
			  QM_RENDER_TRACE_SPLINE,	// Render the C_\alpha trace with a smooth spline
			  QM_RENDER_RIBBON,			// Render the C_\alpha trace with a solid ribbon
										// (based on secondary structure).
			  QM_RENDER_RIBBON_FLAT,	// Render the C_\alpha trace with a flat ribbon
										// (based on secondary structure).
			  QM_RENDER_TUBE,			// Render the C_\alpha trace with a solid tube
			  QM_RENDER_TUBE_T,			// Render the C_\alpha trace with a solid tube
										// of radius T_i (where T_i is the temperature factor
										// of the i^{th} residue).
			  QM_RENDER_TUBE_OCC		// Render the C_\alpha trace with a solid tube
										// of radius OCC_i (where OCC_i is the occupancy
										// of the i^{th} residue).
} RENDER_FORMAT;

#define DEFAULT_PDB_SERVER	"www.rcsb.org"

// Define the boolean constants
#ifndef TRUE
#define TRUE	1
#endif

#ifndef FALSE
#define FALSE	0
#endif

// Define index numbers for the status bar
enum {INFO_PANE = 0, ATOM_NUM_PANE, ATOM_ID_PANE, AMINO_ACID_PANE};

// The default viewport perspective (for zooming ...)
#define		QM_PERSPECTIVE	60.0f

#define		QM_NEAR_CLIP_Z	0.5f
#define		QM_FAR_CLIP_Z	7.0f

#define		QM_DEFAULT_DEPTH	-4.0f

#define 	SCALE_FACTOR 4.0f

#define	TORSION_RADIUS		0.3f
#define	CYLINDER_HEIGHT		0.1f
#define	MEASURE_RADIUS		0.4f
#define	MONITOR_RADIUS		0.15f

#define	STICK_RADIUS		0.175f
#define	BALL_RADIUS			0.275f

// The default maximum size of the bitmap produced
// when an image is printed
#define	QMOL_DEFAULT_MEM_SIZE 8192

// Different rendering modes
enum {QM_RENDER_SCREEN = 0, QM_RENDER_SELECT};

#define		QM_FORWARD		0
#define		QM_REVERSE		1

// Used for rendering 3D objects
enum {SELECTION_SPHERE = 1, ATOM_SPHERE, TORSION_DISK, TORSION_CYL,
BOND_CYL, AXIS_CONE};

// Define different user-selectable resolutions for rendering polygons
enum {HIGH_RES, MED_RES, LOW_RES, CUSTOM_RES};

// The alpha value for blending the spheres representing selected atoms
// with the rest of the scene.
#define	SELECT_ALPHA	150 //128

// Line width for rendering measure monitors
#define		MEASURE_LINE_WIDTH		3.0

// Enumerate the different stero modes
enum {QM_STEREO_NONE, QM_STEREO_SPLIT_SCREEN};

// Maximum size of a DIB (for clipboard and print operations)
// is MAX_DIB_SIZE X MAX_DIB_SIZE. A maximum size is needed because the
// windows API function StretchDIBits(...) apparently fails with larger
// images.
#define		MAX_DIB_SIZE		1000

#endif // __QMOL_DEFINE