//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

// 
/*
 * Last modified 3/21/98 by J. Gans
 *	- CONECT field now stored as a linked list (removed the 
 *	  PDB_MAX_COORDINATION variable)
 * Last modified 1/13/97 by J. Gans
 *	- Converted to c++.
 *	- corrected bug that caused non-ATOM entries to derail the
 *	  parser.
 * Last Modified 3/5/99
 *	- Added a serial to record function.
 * Last Modified 3/31/99
 * - Added support for the HETATM record name (for non-standard groups - 
 *   especially water)
 * - Put all the code in the .h file (for inlining purposes).
 * - Added a Print() function to write the pdb file to stdout.
 * - Added a bounding box
 * Last Modified 4/4/99
 * - Separated the code back into .cpp and .h files (had compiling problems)
 *	Last modified 4/8/99
 *	- Added a bounding box extension (hidden behind a REMARK field).
 *			BBOX_R	...... 	The center of the bounding box
 *			BBOX_A	......	The "x" egde of the box.	
 *			BBOX_B	......	The "y" egde of the box.	
 *			BBOX_C	......	The "z" egde of the box.
 * Last modified 5/16/99
 * - Fixed a bug in counting the number of covalent bonds in the molecule.
 *		see the CountBond() function.
 *	Last modified 6/27/99
 *	- Overloaded the GetCoord() function to also accept double arguments.
 * Last modified 7/16/99
 * - Fixed bug in pdb_record::SetRecord(char *rec). I had forgotten to
 *		write a string terminator at the end of the record string!
 *	Last modified 7/18/99
 *	- Fixed a bug in reading the x, y and z data fields of a pdb record.
 *		There was the possiblility of mapping x->z and z->x.
 *	Last modified 7/19/99
 *	- Fixed a bug in counting the number of bonds.
 *	Last modified 7/26/99
 *	- Fixed a bug in the copy constructor (I had failed to initialize the 
 *		data arrays)
 *	Last modified 7/29/99
 *	- Added the ability to specify the precision of number fields in the
 *		pdb file (NOT PORTABLE!). The data (x, y, z, occ, T) are now stored
 *		as double values. The value of PDB_BUFFER_SIZE has been increased
 *		(to 512) to handle the anticipated increase in line length.
 *
 * Last modified 8/4/99
 *	- Added a function, UpdateConnect, that maintains any connect definitions
 *		that are present in the pdb file while making sure that there is a
 *		complete set of connect records. Added a new water name, "HOH".
 *
 *	Last modified 8/9/99
 *	- Went back to storing x, y and z data as float. Added GetX(), GetY()
 *		and GetZ() functions to access individual degrees of freedom. Made 
 *		some small changes to the UpdateConnect() code to optimize for speed.
 *		Todo: To really get a speed increase when determining the bonded
 *		structure of a molecule, we need to break the molecule into boxes and
 *		perform a grid search (ala VMD or Rasmol).
 *
 *	Last odified 8/12/99
 *	- Added the above mentioned changes to UpdateConnect().
 *		Fixed a bug reading CONECT records (comments at the end of the line
 *		were derailing the parser).
 *
 *	Last modified 8/24/99
 *	- Fixed a bug that caused atom names that started with numbers to 
 *		be assigned the default distance when computing the molecular
 *		connectivity.
 *
 *	Last modified 8/26/99
 *	- Allow small molecule files (or any file for that matter) to skip the
 *		residue name field).
 *
 *	Last modified 10/16/99
 *	-	Fixed bug that confused record parser when reading the chain ID
 *		record from the PDB file.
 *
 *	Last modified 2/28/00
 *	- Added PVM pack and unpack functions
 *	- Replaced the fixed buffer stack (in the pdb_connect class) with a 
 *	  variable length linked list.
 *	- Updated version to 1.1.
 *	
 *	Last modified 3/1/00
 *	- Added binary read() and write() functions. Note that this may
 *	  cause some confusion as there is already a Read() function
 *	  in the pdb_file class.
 *
 *	Last modified 4/26/00 by J. Gans
 *	- Fixed a bug in the caculation of the number of bonds. The
 *	  variable num_bonds did not take into account the bonds
 *	  added or removed by existing CONECT records in the pdb file.
 *
 *	Last modified 1/5/01 by J. Gans
 *	- Fixed bug in pdb constructor (monitor variable was not 
 *	  initialized to NULL).
 *
 *	Last modified 5/16/01 by J.D. Gans
 *	- Added MODEL record to Write(const char* name, int precision).
 *	- Added an option to save (or note save) the CONECT records when
 *	  writing a pdb file.
 *
 *	Last modified 5/25/01 by J.D. Gans
 *	- Added a chainId to the pdb record class.
 *
 * 	Last modified 6/26/01 by J.D. Gans
 *	- Added a use_serial flag to UpdateConnect() and ReConnect() to
 *	provide more flexibilty in the construction of the connectivity
 *	tables.
 *
 *	Last modified 7/5/01 by J.D. Gans
 *	- Fixed bug in ParseRecord() to handle special case when the atom name and
 *	the residue name have been concatinated. Other misc fixes.
 *
 *	Last modified 7/5/01 by J.D. Gans
 *	- More misc fixes for the parsing of funky HETATOM records.
 *
 * 	Last modified 1/30/02 by J.D. Gans
 *	- Fixed a bug that prevented the reading of multiple model files that failed
 *	to specify model numbers (but gave chain ID's instead).
 *
 *	Last modified 2/21/02 by J.D. Gans
 *	- Fixed a bug that failed to properly correct for long atom names (that 
 *	were concatinated with the residue name).
 */

/* The following info is from the Brookhave PDB format specification: 
 *
 * Record Format
 *
 * COLUMNS        DATA TYPE       FIELD         DEFINITION                            
 * 1 - 6          Record name     "ATOM" or "HETATM"                                              
 * 7 - 11         Integer         serial        Atom serial number                    
 * 13 - 16        Atom            name          Atom name                             
 * 17             Character       altLoc        Alternate location indicator          
 * 18 - 20        Residue name    resName       Residue name                          
 * 22             Character       chainID       Chain identifier                      
 * 23 - 26        Integer         resSeq        Residue sequence number               
 * 27             Character       iCode         Code for insertion of residues        
 * 31 - 38        Real(8.3)       x             Orthogonal coordinates for X          
 * 39 - 46        Real(8.3)       y             Orthogonal coordinates for Y          
 * 47 - 54        Real(8.3)       z             Orthogonal coordinates for Z          
 * 55 - 60        Real(6.2)       occupancy     Occupancy                             
 * 61 - 66        Real(6.2)       tempFactor    Temperature factor                    
 * 68 - 70        Integer         ftNote        Footnote number, being deprecated     
 * 73 - 76        LString(4)      segID         Segment identifier, left-justified    
 * 77 - 78        LString(2)      element       Element symbol, right-justified       
 * 79 - 80        LString(2)      charge        Charge on the atom, IUPAC form 
 */

///////////////////////////////////////////////////////////////
// NOTE: The pdb class uses stdio.h, not iostream.h for all io. 
///////////////////////////////////////////////////////////////
#ifndef	__PDB
#define	__PDB

// Some version control
#define		PDB_VERSION		"1.1"

#include <stdio.h>
#include <fstream.h>
#include <string.h>
#include <ctype.h>
#include "list.h"

#ifdef USE_ZLIB
// Use zlib for reading compressed PDB files
#include "zlib.h"
#else
// If we're NOT using zlib for reading compressed files,
// define gzFile to be FILE*.
typedef FILE* gzFile;
#endif

#ifdef PVM3
#include <pvm3.h> 
#endif 

/* Define Boolean values */
#ifndef TRUE
#define TRUE	1
#endif

#ifndef FALSE
#define FALSE	0
#endif

/*
 * The following fields have been OMITTED in my description of the pdb file
 * format:
 *	    Alternate location indicator
 *	    Chain identifier
 *	    Code for insertion of residues
 *	    Footnote number, being deprecated
 *	    Segment identifier, left-justified
 *	    Element symbol, right-justified
 *	    Charge on the atom, IUPAC form
 */

/* Buffer size for reading in lines of data from the PDB file */
#define PDB_BUFFER_SIZE		512

// According to the format specifier, only the first 31 columns should
// contain data - could cause a problem with large integer values.
#define PDB_CONECT_BUFFER_SIZE	31

#define PDB_RECORD_FIELD	    	(6 + 1)
#define PDB_SERIAL_FIELD	    	(5 + 1)
#define PDB_NAME_FIELD		    	(6 + 1)
#define PDB_RESIDUE_FIELD	    	(5 + 1)
#define PDB_SEQUENCE_FIELD	    	(8 + 1)
#define PDB_X_FIELD		    		(8 + 1)
#define PDB_Y_FIELD		    		(8 + 1)
#define PDB_Z_FIELD		    		(8 + 1)
#define PDB_OCCUPANCY_FIELD		(6 + 1)
#define PDB_TEMPATURE_FIELD		(6 + 1)

/*
 * Here a slew of atomic radii (in angstroms) lifted from the 
 * babel program (from the file element.lis). Used to determine
 * connectivity.
 */

// fudge factor...
#define	    RADIUS_SLOP	0.225

#define	    RADIUS_H		0.230
#define	    RADIUS_C		0.680
#define	    RADIUS_N		0.680	    
#define	    RADIUS_O		0.680
#define	    RADIUS_P		0.750
#define	    RADIUS_S		1.020
#define	    RADIUS_NA		0.970	    
#define	    RADIUS_CL		0.990
#define	    RADIUS_MG		1.100
#define	    RADIUS_AL		1.350
#define	    RADIUS_FE		1.340
#define	    RADIUS_ZN		1.450
#define	    RADIUS_SE		1.220

class pdb_record {
	private:
		// Record name (1 - 6)
		char record[PDB_RECORD_FIELD];
	
		// Atom serial number (7 - 11)
		int serial;
	
		// Atom name (12 - 17)
		char name[PDB_NAME_FIELD]; 

		// Residue (18 - 22)
		char resName[PDB_RESIDUE_FIELD];
	
		// Residue sequence number (23 - 30)
		int resSeq;
	
		// x Coordinate (31 - 38)
		float x;
	
		// y coordinate (39 - 46)
		float y;
	
		// z coordinate ( 47 - 54)
		float z;
	
		// Occupancy (55 - 60)
		float occupancy;
	
		// Temperature factor (61 - 66)
		float tempFactor;
	
		// Is this atom part of a water molecule?
		int is_water;

		// The model number - used when specifying multiple copies of
		// a molecule in the same pdb file (for NMR).
		int model;

		// The chain ID for the atom
		char chainId;
	public:
    
		// Provide public access to set the above variables
		void SetRecord(char *rec);
		inline void SetSerial(int num)
		{
			serial = num;
		};
		
		inline void SetModel(int m)
		{
			model = m;
		};

		inline void SetChainId(char c)
		{
			chainId = c;
		};

		void SetAtomName(char *atom);
		void SetResName(char *resname);
		inline void SetResNum(int num)
		{
			resSeq = num;
		};
		
		inline void SetCoord(float a, float b, float c)
		{
			x = a;
    		y = b;
    		z = c;
		};
		
		inline void SetOcc(float occ)
		{
			occupancy = occ;
		};
		
		inline void SetTemp(float temp)
		{
			tempFactor = temp;
		};

		// Provide public access to read the above variables
		inline char* GetRecord()
		{
			return record;
		};
		
		inline int GetSerial()
		{
			return serial;
		};
		
		inline char* GetAtomName()
		{
			return name;
		};
		
		inline char* GetResName()
		{
			return resName;
		};
		
		inline int GetResNum()
		{
			return resSeq;
		};
		
		inline void GetCoord(float &a, float &b, float &c)
		{
			a = x;
    		b = y;
    		c = z;
		};

		inline void GetCoord(double &a, double &b, double &c)
		{
			a = x;
    		b = y;
    		c = z;
		};
		
		inline float GetX()
		{
			return x;
		};
		
		inline float GetY()
		{
			return y;
		};

		inline float GetZ()
		{
			return z;
		};
		
		inline float GetOcc()
		{
			return occupancy;
		};
		
		inline float GetTemp()
		{
			return tempFactor;
		};
		
		inline int GetModel()
		{
			return model;
		};

		inline char GetChainId()
		{
			return chainId;
		};

		pdb_record& operator =(const pdb_record& rec);
		int ParseRecord(int &m, pdb_record* ptr_last_record, gzFile f_in);
		
		int IsWater(){return is_water;};
		
		int IsDummy()
		{
			if(name[0] == 'D')
				return TRUE;

			return FALSE;
		};
};

struct pdb_bond{
	int a, b;
};

class pdb_connect{
	private:
		int node; // The reference atom
		list<int> branch;
	public:	
		// Different constructors
		pdb_connect(); // Default
		pdb_connect(pdb_connect& source); // Copy
	
		inline int GetNumConnect()
		{
			return branch.Length();
		};
		
		inline int GetNode()
		{
			return node;
		};
		
		inline void SetNode(int n)
		{
			node = n;
		};
	
		// Add a connection
		inline void Add(int n)
		{
			branch.Push(n);
		};
		
		// Remove a connection
		void Remove(int n)
		{
			for(int i = 0;i < branch.Length();i++){
				if(branch[i] == n){
					branch.Remove(i);
					return;
				}
			}

			// If the connection is not found, do nothing!
		};

		int ParseConnect(gzFile fin);
		void Write(FILE* fout);
		inline pdb_connect& operator =(pdb_connect& conn)
		{
			node = conn.node;
    		branch = conn.branch;
	
    		return (*this);
		};
		
		inline int& operator [](int index)
		{
			#ifdef PDB_DEBUG
			if((index < 0) || (index >= branch.GetLength()))
				QMFatalError("pdb_connect: Index out of bounds in pdb_connect");
			#endif //PDB_DEBUG
	
			return branch[index];
		};
};

class pdb_monitor {
	public:
		unsigned int color[3]; // R, G, B: where 256 > color[i] >= 0
		int atom_serial[2]; // serial numbers (as defined in the pdb file)
		int atom_index[2]; // Atom index numbers corresponding to the
						   // above serial numbers.
		int pairId;	// A user defined pairId

		int ParseMonitor(gzFile fin);
};

class pdb_file {
	private:
		int num_bonds; // The number of covalent bonds
					   // (not equal to the number of connect records).
		
		int num_records; // Number of records
		int num_connect; // Number of CONECT entries
		int num_monitors; // Number of monitors

		pdb_record* line; // a pdb file is an array of pdb records.
		pdb_connect* conn; // An array of connection entries
		pdb_monitor* monitor; // An array of monitor entries

		// Is there a bounding
		int bbox;
		float bbox_R[3], bbox_A[3], bbox_B[3], bbox_C[3];
		
		int GetBBox(gzFile f_in);
		int CountMonitors(gzFile f_in);
		int CountRecords(gzFile f_in);
		int CountConnect(gzFile f_in);
		int CountBonds();

		inline void CheckBounds(int record, int line_num)
		{
			if((record >= num_records) || (record < 0))
				throw QmolError("CheckBounds: Index out of bounds.", __FILE__, __LINE__);
		};
		
	public:
		~pdb_file(); // Destructor
		pdb_file(); // Default constructor
		pdb_file(int records); // new pdb file constructor
		pdb_file(const char *filename); // pdb file constructor
		pdb_file(const pdb_file& pdb); // Copy constructor

		inline int GetNumRecords()
		{
			return num_records;
		};
		
		inline int GetNumConnections()
		{
			return num_connect;
		};
		
		inline int GetNumMonitors()
		{
			return num_monitors;
		};

		int GetNumBonds()
		{
			return num_bonds;
		};
		
		// Is there a bounding box
		int IsBBox(){return bbox;};
		
		void GetBBoxR(float &x, float &y, float &z);
		void GetBBoxA(float &x, float &y, float &z);
		void GetBBoxB(float &x, float &y, float &z);
		void GetBBoxC(float &x, float &y, float &z);
		
		inline char* GetRecord(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
			return line[record].GetRecord();
		};
		
		inline pdb_connect& GetConnect(int record)
		{
			#ifdef PDB_DEBUG
			if((record < 0) || (record >= num_connect))
				QMFatalError("GetConnect: Invalid index value!");
 			#endif //PDB_DEBUG
	 
   			return conn[record];
		};
		
		inline pdb_monitor& GetMonitor(int record)
		{
			#ifdef PDB_DEBUG
			if((record < 0) || (record >= num_monitors))
				QMFatalError("GetMonitor: Invalid index value!");
 			#endif //PDB_DEBUG
	 
   			return monitor[record];
		};

		void ResetConnect(int m_num_connect = -1)
		{
			if(conn)
				delete [] conn;

			if(m_num_connect < 0)
				num_connect = num_records;
			else
				num_connect = m_num_connect;

			conn = new pdb_connect [num_connect];
			for(int i = 0;i < num_connect;i++)
				conn[i].SetNode(line[i].GetSerial());
		};
		
		void PushConnect(int record, int i)
		{
			#ifdef PDB_DEBUG
			if((record < 0) || (record >= num_connect))
				QMFatalError("PushConnect: Invalid index value!");
 			#endif //PDB_DEBUG
			
			conn[record].Add(i);
		}
		
		// Is the selected atom part of a water molecule?
		int IsWater(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].IsWater();
		};
		
		
		// Is the specified atom a dummy atom?
		int IsDummy(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].IsDummy();
		};
		
		int GetSerial(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG

			return line[record].GetSerial();
		};
		
		char* GetAtomName(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].GetAtomName();
		};
		
		char* GetResName(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
   
			return line[record].GetResName();
		};
		
		int GetResNum(int record)
		{
			 #ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].GetResNum();
		};
		
		void GetCoord(int record, float &a, float &b, float &c)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].GetCoord(a, b, c);
		};
		
		void GetCoord(int record, double &a, double &b, double &c)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].GetCoord(a, b, c);
		};
		
		float GetX(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].GetX();
		};
		
		float GetY(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].GetY();
		};
		
		float GetZ(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].GetZ();
		};
		
		float GetOcc(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].GetOcc();
		};
		
		float GetTemp(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].GetTemp();
		};

		int GetModel(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].GetModel();
		};

		int GetChainId(int record)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			return line[record].GetChainId();
		};

		void SetRecord(int record, char *rec)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetRecord(rec);
		};
		
		void SetSerial(int record, int num)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetSerial(num);
		};
		
		void SetAtomName(int record, char *atom)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetAtomName(atom);
		};
		
		void SetResName(int record, char *resname)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetResName(resname);
		};
		
		void SetResNum(int record, int num)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
  	 		line[record].SetResNum(num);
		};
		
		void SetCoord(int record, double a, double b, double c)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetCoord((float)(a), (float)(b), (float)(c));
		};
		
		void SetCoord(int record, float a, float b, float c)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetCoord(a, b, c);
		};
		
		void SetOcc(int record, double occ)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetOcc((float)(occ));
		};
		
		void SetOcc(int record, float occ)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetOcc(occ);
		};
		
		void SetTemp(int record, double temp)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetTemp((float)(temp));
		};
		
		void SetTemp(int record, float temp)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetTemp(temp);
		};
		
		void SetModel(int record, int model)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetModel(model);
		};
		
		void SetChainId(int record, char c)
		{
			#ifdef PDB_DEBUG
			CheckBounds(record, __LINE__);
			#endif //PDB_DEBUG
	
   			line[record].SetChainId(c);
		};

		pdb_file& operator =(const pdb_file& pdb);
		
		// Get the coordinates by the atom serial number, not
		// array index of the pdb_record.
		void GetCoordBySerial(int num, double &a, double &b, 
							  double &c);
		
		// Given the serial number, return the line number.
		int HashSerial(int &rec, int serial);

		// Concatenate the two files a & b
		void Concatinate(const pdb_file& a, const pdb_file& b);
		void Translate(float *r);
		void Translate(double *r);
		void Rotate(float **m);
		void Rotate(double **m);
		int Read(const char *file);
		
		// Allow the user to specify the precision of the 
		// coordinate fields in the output file. Note that '3'
		// is the default precision
		int Write(const char *file, int precision = 3, BOOL write_conect = TRUE);
		int Write(FILE* fout, int precision = 3, BOOL write_conect = TRUE);

		void Print(BOOL write_conect = TRUE);
		
		// Compute the molecular connectivity from scratch.
		void ReConnect(int use_serial = TRUE);
		
		// Suppliment any molecular connectivity that is currently
		// part of the pdb file such that all of the connection 
		// records are defined. Return the value of use_serial.
		int UpdateConnect(int use_serial = TRUE);
		
		// Get the van der waals radius
		float GetVDW(int i)
		{
			int j = 0;
			
			#ifdef PDB_DEBUG
			CheckBounds(i, __LINE__);
			#endif //PDB_DEBUG
			
			// Check all of the letters in the atom name untill
			// we find a non-numeric character. People occasionally
			// place numbers in front of hydrogen atom names (to
			// differentiate between the various H's).
			char c = line[i].GetAtomName()[0];
			while(!isalpha(c) && (line[i].GetAtomName()[j+1] != '\0'))
				c = line[i].GetAtomName()[++j];
	
			switch(c){
				case 'C': // Carbon
					return (float)(RADIUS_C + RADIUS_SLOP);
					break;
				case 'H': // hydrogen
					return (float)(RADIUS_H + RADIUS_SLOP);
					break;
				case 'N': // Nitrogen
					return (float)(RADIUS_N + RADIUS_SLOP);
					break;
				case 'O': // Oxygen
					return (float)(RADIUS_O + RADIUS_SLOP);
					break;
				case 'S': // Sulfur
					return (float)(RADIUS_S + RADIUS_SLOP);
					break;
				case 'P': // Phosphorous
					return (float)(RADIUS_P + RADIUS_SLOP);
					break;
				case 'D': // Dummy atom - pulled radius out of thin air
					return (float)(RADIUS_C);
					break;
				default:
					return (float)(RADIUS_C + RADIUS_SLOP);
					break;
			}
		};
		
		// Return the record number associated with the given serial
		// number
		int SerialToRecord(int i);

		// Get the largest and smallest values of the temperature and occupancy
		float GetMaxTemp();
		float GetMinTemp();
		float GetMaxOcc();
		float GetMinOcc();
};

#endif /* __PDB */
