//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
// 
//
// A template stack implementation using a fixed stack buffer.
// 4/1/99 J. Gans
//

#ifndef __STACK
#define	__STACK

// Define STACK_DEBUG to turn on bounds testing
//#define STACK_DEBUG

#include "Excepts.h"

#ifdef PVM3
#include <pvm3.h>
#endif // PVM3

template <class A>
class stack{
	private:
		int length;
		int max;
		A *buff;
	public:
		stack(int m = 0);
		~stack();
		
		void Allocate(int m);
		
		int GetLength(){ return length;};
		void Push(const A& item);
		A Pop();
		A& operator[](int i);
		stack& operator=(stack<A>& s);

		#ifdef PVM3
		// Rely on the base type declarations of pack and
		// unpack to handle packing and unpacking.
		void pack()
		{
			QMFatalError("Unknown data type in pack!");
		};

		void unpack()
		{
			QMFatalError("Unknown data type in unpack!");
		};
		#endif // PVM3
};

template <class A>
stack<A>::stack(int m)
{
	ASSERT(m >= 0);

	if(m == 0)
		buff = NULL;
	else
		buff = new A [m];
	
	max = m;
	length = 0;
}

template <class A>
stack<A>::~stack()
{
	if(buff)
		delete [] buff;
}

template <class A>
inline void stack<A>::Push(const A& item)
{
	#ifdef STACK_DEBUG
	if(length == max)
		throw QMOL_ERROR("Maximum buffer size exceeded in Push!");
	#endif //STACK_DEBUG
	
	buff[length++] = item;	
}

template <class A>
inline A stack<A>::Pop()
{
	#ifdef STACK_DEBUG
	if(length == 0)
		throw QMOL_ERROR("Can't Pop empty stack!");
	#endif //STACK_DEBUG
	
	length --;
	
	return buff[length];
}


// Reallocate (and resize) the stack buffer
template <class A>
inline void stack<A>::Allocate(int i)
{
	if(buff)
		delete [] buff;
	
	buff = new A [i];
	
	length = 0;
	max = i;
}


template <class A>
inline A& stack<A>::operator[](int i)
{
	#ifdef STACK_DEBUG
	if((i < 0) || (i >= length))
		throw QMOL_ERROR("Index out of bounds in []");
	#endif // STACK_DEBUG
	
	return buff[i];
}

template <class A>
inline stack<A>& stack<A>::operator=(stack<A>& s)
{
	length = s.length;
	max = s.max;

	if(buff)
		delete [] buff;
	
	if(max > 0){
		buff = new A [max];
		
		for(int i = 0;i < length;i++)
			buff[i] = s.buff[i];
	}
	else{
		buff = NULL;
	}
		
	return (*this);
}

#ifdef PVM3

// Handle various base types explicitly
void stack<int>::pack()
{
	// Pack the current length
	pvm_pkint(&length, 1, 1);
	
	// Pack the maximum stack size
	pvm_pkint(&max, 1, 1);
	
	// Pack the stack
	pvm_pkint(buff, length, 1);
}

void stack<float>::pack()
{
	// Pack the current length
	pvm_pkint(&length, 1, 1);
	
	// Pack the maximum stack size
	pvm_pkint(&max, 1, 1);
	
	// Pack the stack
	pvm_pkfloat(buff, length, 1);
}

void stack<double>::pack()
{
	// Pack the current length
	pvm_pkint(&length, 1, 1);
	
	// Pack the maximum stack size
	pvm_pkint(&max, 1, 1);
	
	// Pack the stack
	pvm_pkdouble(buff, length, 1);
}

void stack<char>::pack()
{
	// Pack the current length
	pvm_pkint(&length, 1, 1);
	
	// Pack the maximum stack size
	pvm_pkint(&max, 1, 1);
	
	// Pack the stack
	pvm_pkbyte(buff, length, 1);
}

void stack<int>::unpack()
{
	int tmp;

	// Unpack the current length
	pvm_upkint(&tmp, 1, 1);
	
	// Unpack the maximum stack size
	pvm_upkint(&max, 1, 1);
	
	// Allocate room for the stack (which also 
	// sets the length to 0).
	Allocate(max);
	length = tmp;

	// Unpack the stack
	pvm_upkint(buff, length, 1);
}

void stack<float>::unpack()
{
	int tmp;

	// Unpack the current length
	pvm_upkint(&tmp, 1, 1);
	
	// Unpack the maximum stack size
	pvm_upkint(&max, 1, 1);
	
	// Allocate room for the stack (which also 
	// sets the length to 0).
	Allocate(max);
	length = tmp;

	// Unpack the stack
	pvm_upkfloat(buff, length, 1);
}

void stack<double>::unpack()
{
	int tmp;

	// Unpack the current length
	pvm_upkint(&tmp, 1, 1);
	
	// Unpack the maximum stack size
	pvm_upkint(&max, 1, 1);
	
	// Allocate room for the stack (which also 
	// sets the length to 0).
	Allocate(max);
	length = tmp;

	// Unpack the stack
	pvm_upkdouble(buff, length, 1);
}

void stack<char>::unpack()
{
	int tmp;

	// Unpack the current length
	pvm_upkint(&tmp, 1, 1);
	
	// Unpack the maximum stack size
	pvm_upkint(&max, 1, 1);
	
	// Allocate room for the stack (which also 
	// sets the length to 0).
	Allocate(max);
	length = tmp;

	// Unpack the stack
	pvm_upkbyte(buff, length, 1);
}


#endif // PVM3

#endif // __STACK
