//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

#include "stdafx.h"

#include "trackball.h"
#include "LinearAlgebra3D.h"
#include <math.h>

VirtualTrackBall::VirtualTrackBall()
{
	curquat[0] = lastquat[0] = 0.0f;
	curquat[1] = lastquat[1] = 0.0f;
	curquat[2] = lastquat[2] = 0.0f;
	curquat[3] = lastquat[3] = 1.0f;

	// The default value
	trackballsize = DEFAULT_TRACKBALLSIZE;
}

void VirtualTrackBall::trackball(float p1x, float p1y, 
								 float p2x, float p2y)
{
	// Axis of rotation.
	static float a[3]; 

	// How much to rotate about axis.
	float phi;            
	static float p1[3], p2[3], d[3];
	float t;

	if (p1x == p2x && p1y == p2y) {
		// Zero rotation
		lastquat[0] = 0.0f;
		lastquat[1] = 0.0f;
		lastquat[2] = 0.0f;
		lastquat[3] = 1.0f;
		return;
	}

	// First, figure out z-coordinates for projection of P1 and
	// P2 to deformed sphere.
	p1[0] = p1x;
	p1[1] = p1y;
	p1[2] = tb_project_to_sphere(trackballsize, p1x, p1y);
	
	p2[0] = p2x;
	p2[1] = p2y;
	p2[2] = tb_project_to_sphere(trackballsize, p2x, p2y);

	// Now, we want the cross product of P1 and P2.
	//a = p2 | p1; // a = p2 X p1
	CROSS(a, p2, p1);

	// Figure out how much to rotate around that axis. 
	d[X] = p1[X] - p2[X];
	d[Y] = p1[Y] - p2[Y];
	d[Z] = p1[Z] - p2[Z];

	t = (float)(sqrt(DOT(d,d))/(2.0f * trackballsize));

	// Avoid problems with out-of-control values.
	if (t > 1.0f)
		t = 1.0f;
	if (t < -1.0f)
		t = -1.0f;
	phi = (float)(2.0 * asin(t));

	axis_to_quat(a, phi);
}

// Given an axis and angle, compute quaternion.
void VirtualTrackBall::axis_to_quat(float a[3], float phi)
{
	// Normalize a
	float tmp = 1.0f/(float)sqrt(DOT(a, a));

	a[X] *= tmp;
	a[Y] *= tmp;
	a[Z] *= tmp;

	lastquat[0] = a[0];
	lastquat[1] = a[1];
	lastquat[2] = a[2];
	
	tmp = (float)(sin(phi/2.0f));

	lastquat[X] *= tmp;
	lastquat[Y] *= tmp;
	lastquat[Z] *= tmp;
	
	lastquat[3] = (float)(cos(phi/2.0f));
}

// Project an x,y pair onto a sphere of radius r OR a
// hyperbolic sheet if we are away from the center of the
// sphere.
float VirtualTrackBall::tb_project_to_sphere(float r, float x, float y)
{
	float d, t, z;

	d = (float)(sqrt(x * x + y * y));
	if (d < r * 0.70710678118654752440f) {
		// Inside sphere
		z = (float)(sqrt(r * r - d * d));
	} 
	else {
		// On hyperbola
		t = r / 1.41421356237309504880f;
		z = t * t / d;
	}
	
	return z;
}

void VirtualTrackBall::update()
{
	static int count = 0;
	static float t1[4], t2[4], t3[4], tf[4];
	float three_d_dot;

	t1[0] = lastquat[0]*curquat[3];
	t1[1] = lastquat[1]*curquat[3];
	t1[2] = lastquat[2]*curquat[3];

	t2[0] = curquat[0]*lastquat[3];
	t2[1] = curquat[1]*lastquat[3];
	t2[2] = curquat[2]*lastquat[3];
	
	// Cross Product
	t3[0] = curquat[1]*lastquat[2] - curquat[2]*lastquat[1];
	t3[1] = curquat[2]*lastquat[0] - curquat[0]*lastquat[2];
	t3[2] = curquat[0]*lastquat[1] - curquat[1]*lastquat[0];
	
	tf[0] = t1[0] + t2[0] + t3[0];
	tf[1] = t1[1] + t2[1] + t3[1];
	tf[2] = t1[2] + t2[2] + t3[2];

	// Take the dot product in 3D
	three_d_dot = DOT(lastquat, curquat);

	tf[3] = lastquat[3] * curquat[3] - three_d_dot;

	curquat[0] = tf[0];
	curquat[1] = tf[1];
	curquat[2] = tf[2];
	curquat[3] = tf[3];

	if (++count > RENORMCOUNT) {
		count = 0;

		float mag = 1.0f/(curquat[0]*curquat[0] + curquat[1]*curquat[1] +
					curquat[2]*curquat[2] + curquat[3]*curquat[3]);

		curquat[0] *= mag;
		curquat[1] *= mag;
		curquat[2] *= mag;
		curquat[3] *= mag;
	}
}


/* Build a rotation matrix, given a quaternion rotation. */
void VirtualTrackBall::build_rotmatrix(GLfloat m[4][4])
{
	m[0][0] = 1.0f - 2.0f * (curquat[1] * curquat[1] + 
							 curquat[2] * curquat[2]);
	m[0][1] = 2.0f * (curquat[0] * curquat[1] - curquat[2] * curquat[3]);
	m[0][2] = 2.0f * (curquat[2] * curquat[0] + curquat[1] * curquat[3]);
	m[0][3] = 0.0f;

	m[1][0] = 2.0f * (curquat[0] * curquat[1] + curquat[2] * curquat[3]);
	m[1][1] = 1.0f - 2.0f * (curquat[2] * curquat[2] + 
							 curquat[0] * curquat[0]);
	m[1][2] = 2.0f * (curquat[1] * curquat[2] - curquat[0] * curquat[3]);
	m[1][3] = 0.0f;

	m[2][0] = 2.0f * (curquat[2] * curquat[0] - curquat[1] * curquat[3]);
	m[2][1] = 2.0f * (curquat[1] * curquat[2] + curquat[0] * curquat[3]);
	m[2][2] = 1.0f - 2.0f * (curquat[1] * curquat[1] + 
							 curquat[0] * curquat[0]);
	m[2][3] = 0.0f;

	m[3][0] = 0.0f;
	m[3][1] = 0.0f;
	m[3][2] = 0.0f;
	m[3][3] = 1.0f;
}
